/*
    Copyright (C) 2002 Paul Davis

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#ifndef __jack_timestamps_h__
#define __jack_timestamps_h__

#include <stdio.h>

#ifdef __cplusplus
extern "C"
{
#endif

    void jack_init_timestamps (unsigned long howmany);
    void jack_timestamp (const char *what);
    void jack_dump_timestamps (FILE *out);
    void jack_reset_timestamps ();

#ifdef __cplusplus
}
#endif

#endif /* __jack_timestamps_h__ */


