// Generated by ResEdit 1.4.3
// Copyright (C) 2006-2008
// http://www.resedit.net

#include "resource.h"
#include "afxres.h"


//
// Version Information resources
//
LANGUAGE LANG_FRENCH, SUBLANG_DEFAULT
1 VERSIONINFO
    FILEVERSION     1,9,11,0
    PRODUCTVERSION  1,9,11,0
    FILEOS          VOS_UNKNOWN
    FILETYPE        VFT_DLL
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040c04b0"
        BEGIN
            VALUE "Comments", "\0"
            VALUE "CompanyName", "Grame\0"
            VALUE "FileDescription", "Jackmp Audio Adapter for Windows\0"
            VALUE "FileVersion", "1, 9, 11\0"
            VALUE "InternalName", "audioadapter\0"
            VALUE "LegalCopyright", "Copyright Grame © 2006-2014\0"
            VALUE "LegalTrademarks", "\0"
            VALUE "OriginalFilename", "audioadapter.dll\0"
            VALUE "PrivateBuild", "\0"
            VALUE "ProductName", "audioadapter\0"
            VALUE "ProductVersion", "1, 9, 11\0"
            VALUE "SpecialBuild", "\0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 1036, 1200
    END
END
